/// This file is part of the X3DToolKit library
/// Copyright (C) 2002-2003  Yannick Le Goc (legoc@imag.fr)

/// http://www-artis.imag.fr/Members/Yannick.Legoc/X3D/

/// This library is free software; you can redistribute it and/or
/// modify it under the terms of the GNU Lesser General Public
/// License as published by the Free Software Foundation; either
/// version 2.1 of the License, or (at your option) any later version.

/// This library is distributed in the hope that it will be useful,
/// but WITHOUT ANY WARRANTY; without even the implied warranty of
/// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
/// Lesser General Public License for more details.

/// You should have received a copy of the GNU Lesser General Public
/// License along with this library; if not, write to the Free Software
/// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
////////////////////////////////////////////////////////////////////////////////
//                            SFTemplateMesh.h                                //
////////////////////////////////////////////////////////////////////////////////

#ifndef TEMPLATEMESH_H
#define TEMPLATEMESH_H

#include "MESH_X3DMeshNode.h"
#include "MESH_SFTemplateVertex.h"
#include "MESH_TemplateVertexSet.h"
#include "MESH_SFTemplateDirectEdge.h"
#include "MESH_SFTemplateIndirectEdge.h"
#include "MESH_SFTemplateFace.h"

#include <iostream>
#include <typeinfo>

namespace X3DTK {
namespace MESH {

/// Template class providing a mesh structure based upon oriented edges, vertices and faces.
/// The integrity of the structure is ensured by the interface of creation. When a new face is 
/// created, the vertices and edges are updated. 

template<class MData, class VData, class EData, class FData>
class TemplateMesh : public X3DMeshNode
{
public:
  /// Defines the Vertex type for the Mesh.
  typedef SFTemplateVertex<VData, EData, FData> SFVertex;
  /// Defines the MVertex type for the Mesh.
  typedef MFTemplateVertex<VData, EData, FData> MFVertex;
  /// Defines the Edge type for the Mesh.  
  typedef SFTemplateEdge<EData, FData, VData> SFEdge;
  /// Defines the MEdge type for the Mesh.  
  typedef MFTemplateEdge<EData, FData, VData> MFEdge;
  /// Defines the Face type for the Mesh.  
  typedef SFTemplateFace<FData, VData, EData> SFFace;
  /// Defines the MFace type for the Mesh.  
  typedef MFTemplateFace<FData, VData, EData> MFFace;
  
  /// Constructor.
  TemplateMesh()
  : X3DMeshNode()
  {
    defineTypeName(typeid(*this).name());
  };
  /// Clones the node.
  virtual SFNode clone() const
  {
    return new TemplateMesh(*this);  
  };
  
  /// Virtual destructor.
  virtual ~TemplateMesh() 
  {
    removeParentFromChild(this, vertexSet);   
  
    // delete edges
    for (typename MFEdge::iterator it = _edges.begin(); it != _edges.end(); ++it)
    {
      delete (*it)->_symetric;
      delete (*it)->_edgeContent;
      delete (*it);    
    } 
     
    // delete faces
    for (typename MFFace::iterator it = _faces.begin(); it != _faces.end(); ++it)
      delete (*it); 
  };
   
  /// Sets the TemplateVertexSet attribute.
  void setVertexSet(TemplateVertexSet<VData, EData, FData> *vertexSet)
  {
    removeParentFromChild(this, this->vertexSet);  
    this->vertexSet = vertexSet;
    addParentToChild(this, this->vertexSet);  
  };
  /// Gets the children list.
  virtual MFAbstractNode getChildrenList() const
  {
    MFAbstractNode childrenList;
    if (vertexSet != 0)
      childrenList.push_back(vertexSet);
    
    return childrenList;   
  };
  
  /// Adds the child N. Returns FALSE if the child already exists or N is not a 
  /// valid child.   
  bool addChild(const SFAbstractNode &N)
  {
    if (dynamic_cast<TemplateVertexSet<VData, EData, FData> *>(N) != 0)
    {
      if (vertexSet != 0)
        return false;
      
      setVertexSet(static_cast<TemplateVertexSet<VData, EData, FData> *>(N));
      return true;
    }
    if (N != 0)
      cx3d << "MESH::TemplateMesh::addChild: a node of type " << N->getTypeName() << " cannot be a child!" << std::endl;
    else
      cx3d << "MESH::TemplateMesh::addChild: cannot add a NULL node!" << std::endl;
   
    return false;
  };
  /// Sets the child N. Returns FALSE if N is not a 
  /// valid child. If the child already exists then N becomes the new child.
  bool setChild(const SFAbstractNode &N)
  {
    if (dynamic_cast<TemplateVertexSet<VData, EData, FData> *>(N) != 0)
    {
      setVertexSet(static_cast<TemplateVertexSet<VData, EData, FData> *>(N));
      return true;
    }
    if (N != 0)
      cx3d << "MESH::TemplateMesh::setChild: a node of type " << N->getTypeName() << " cannot be a child!" << std::endl;
    else
      cx3d << "MESH::TemplateMesh::setChild: cannot add a NULL node!" << std::endl;
    
    return false;
  };
  
  /// Removes the child if exists. Returns TRUE if succeeds.
  bool removeChild(const SFAbstractNode &N)
  {
    if (vertexSet == N)
    {
      setVertexSet(0);
      return true;
    }
    
    return false;
  };    
  
  /// Get the edge from - to.
  SFEdge *getEdge(SFVertex *from, SFVertex *to)
  {
    if ((from == 0) || (to == 0))
      return 0;
    
    MFEdge from_edges = from->getEdges();
    
    for (typename MFEdge::const_iterator it = from_edges.begin(); it != from_edges.end(); ++it)
    {
      if ((*it)->getToVertex() == to) 
        return (*it);
    }
        
    return 0;
  };
  
  /// Gets the Vertices that can be shared with other Mesh.
  const MFVertex &getVertices() const 
  {
    if (vertexSet != 0)
      return static_cast<TemplateVertexSet<VData, EData, FData> *>(vertexSet)->getVertices();
    
    cx3d << "MESH::TemplateMesh::getVertices: cannot return vertices because VertexSet is NULL!" << std::endl;  
    static MFVertex mv;
    return mv;
  };
    
  /// Gets the edges.
  inline const MFEdge &getEdges() const {return _edges;};
  /// Gets the faces.
  inline const MFFace &getFaces() const {return _faces;};
  /// Gets the data.  
  inline MData &data() {return _data;};
  /// Gets the data.  
  inline const MData &data() const {return _data;};
  
  /// Creates Face at the end.
  SFFace *createFace(const MFInt32 &coordIndex, bool ccw = true) 
  {
    return createFace(_faces.size(), coordIndex, ccw);
  };
    
  /// Creates a new face at position i.
  SFFace *createFace(SFInt32 i, const MFInt32 &coordIndex, bool ccw = true) 
  {
    // Creating the Face.
    if (static_cast<unsigned int>(i) >= _faces.size())
      _faces.resize(i + 1, 0);
    else
    {
      if (_faces[i] != 0)  
      {
        cx3d << "TemplateMesh::createFace: cannot create a Face at index " << i << ", a Face already exists!" << std::endl;
        return 0;
      }
    }
      
    SFFace *F = new SFFace(i);
    _faces[i] = F;
    
    if (ccw)
    {
      // Creating the Edges.
      MFInt32::const_iterator itCoordIndex = coordIndex.begin();
      SFVertex *A, *B;
      SFEdge *E;
    
      const MFVertex &mv = getVertices();
      SFVertex *start = mv[*itCoordIndex]; // to test...
    
      A = start;
      ++itCoordIndex;
    
      while (itCoordIndex != coordIndex.end())
      {
        //B = getVertices[*itCoordIndex];
        B = mv[*itCoordIndex];
        ++itCoordIndex;
      
        E = getEdge(A, B);
        if (E == 0)
          E = createEdge(A, B);
        
        E->addLeftFace(F);    
        E->getFromVertex()->addFace(F);    
        F->addEdge(E);
      
        A = B;
      }
      B  = start;
      E = getEdge(A, B);
      if (E == 0)
        E = createEdge(A, B);
    
      E->addLeftFace(F);    
      E->getFromVertex()->addFace(F);
      F->addEdge(E);
    }
    else
    {
      // Creating the Edges.
      MFInt32::const_reverse_iterator itCoordIndex = coordIndex.rbegin();
      SFVertex *A, *B;
      SFEdge *E;
    
      const MFVertex &mv = getVertices();
      SFVertex *start = mv[*itCoordIndex]; // to test...
    
      A = start;
      ++itCoordIndex;
    
      while (itCoordIndex != coordIndex.rend())
      {
        //B = getVertices[*itCoordIndex];
        B = mv[*itCoordIndex];
        ++itCoordIndex;
      
        E = getEdge(A, B);
        if (E == 0)
          E = createEdge(A, B);
        
        E->addLeftFace(F);    
        E->getFromVertex()->addFace(F);    
        F->addEdge(E);
      
        A = B;
      }
      B = start;
      E = getEdge(A, B);
      if (E == 0)
        E = createEdge(A, B);
    
      E->addLeftFace(F);    
      E->getFromVertex()->addFace(F);
      F->addEdge(E);
    }
            
    return F;
  };
  
protected:
  /// Copy constructor.
  TemplateMesh(const TemplateMesh &N)
  : X3DMeshNode(N), _edges(N._edges), _faces(N._faces), _data(N._data)
  {
    addChild(N.vertexSet);
  };
  
private:
  MFEdge _edges;
  MFFace _faces;
  MData _data;
  
  /// Creates a new edge.
  SFEdge *createEdge(SFVertex *from, SFVertex *to) 
  {
    if ((from == 0) || (to == 0))
      return 0;
   
    SFTemplateEdgeContent<EData, FData, VData> *content = new SFTemplateEdgeContent<EData, FData, VData>(from, to);
    SFEdge *E1 = new SFTemplateDirectEdge<EData, FData, VData>(content);
    SFEdge *E2 = new SFTemplateIndirectEdge<EData, FData, VData>(content);
    
    E1->setSymetric(E2);
    E2->setSymetric(E1);
       
    _edges.push_back(E1);
    
    // updating content of A and B.
    // to complete... to respect order of edges.
    from->addEdge(E1);
    to->addEdge(E2);
    
    return E1;
  };
};

}
}

#endif
