/// This file is part of the X3DToolKit library
/// Copyright (C) 2002-2003  Yannick Le Goc (legoc@imag.fr)

/// http://www-artis.imag.fr/Members/Yannick.Legoc/X3D/

/// This library is free software; you can redistribute it and/or
/// modify it under the terms of the GNU Lesser General Public
/// License as published by the Free Software Foundation; either
/// version 2.1 of the License, or (at your option) any later version.

/// This library is distributed in the hope that it will be useful,
/// but WITHOUT ANY WARRANTY; without even the implied warranty of
/// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
/// Lesser General Public License for more details.

/// You should have received a copy of the GNU Lesser General Public
/// License along with this library; if not, write to the Free Software
/// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
///////////////////////////////////////////////////////////////////////////////
//                            X3DComponentVisitor.h                           //
////////////////////////////////////////////////////////////////////////////////

#ifndef X3DCOMPONENTVISITOR_H
#define X3DCOMPONENTVISITOR_H

#include "X3DComponent.h"
#include "GraphTraversal.h"
#include "EnterFunction.h"
#include "WalkOnFunction.h"
#include "LeaveFunction.h"
#include "X3DAbstractNode.h"
#include "SFSceneGraph.h"

namespace X3DTK {

/// Abstract class defining the concept of component visitor.

class X3DTK_API X3DComponentVisitor : public X3DComponent
{
public:
  /// Constructor.
  X3DComponentVisitor();
  /// Destructor.
  virtual ~X3DComponentVisitor() = 0;
  
  /// Gets the enter function of name.
  EnterFunction *getEnterFunctionOf(const SFType *type) const;
  /// Gets the WalkOn function of name.
  WalkOnFunction *getWalkOnFunctionOf(const SFType *type) const;
  /// Gets the leave function of name.
  LeaveFunction *getLeaveFunctionOf(const SFType *type) const;
  
protected:    
  /// Define a new enter function.
  template<class C, class T> 
  void defineEnterFunction(void (C::*ptrF)(T *) const)
  {
    SFType::beginFunctionDefinition();
    EnterFunction *EF = new EnterFunction(reinterpret_cast<ptrToEnterFunction>(ptrF), this);
    //finding if an occurence of T
    T *t = new T();
    SFString name = t->getTypeName();
    SFComponent *TComponent = t->getType()->getComponent();
    //checking if the node belongs to the component.
    if (TComponent != component)
    {
      SFString sg1, sg2;
      if (component->getSceneGraphName() != "")
        sg1 = " of the " + component->getSceneGraphName() + " scene graph";
      if (t->getType()->getSceneGraphName() != "")
        sg1 = " of the " + t->getType()->getSceneGraphName() + " scene graph";  

      cx3d << "warning: defineEnterFunction for " << name << " which doesn't belong to the " << component->getName() << " component" << sg1 << "," << std::endl;
      cx3d << "         but to the " << t->getType()->getComponentName() << " component" << sg2 << std::endl;
    }
    else
      _enterDict.insert(std::pair<SFString, EnterFunction *>(t->getType()->getEncodedName(), EF));

    delete t;
    SFType::endFunctionDefinition();
  };
  
  /// Define a new walk On function.
  template<class C, class T> 
  void defineWalkOnFunction(bool (C::*ptrF)(T *, SFAbstractNode) const)
  {
    SFType::beginFunctionDefinition();
    WalkOnFunction *WF = new WalkOnFunction(reinterpret_cast<ptrToWalkOnFunction>(ptrF), this);
    //finding if an occurence of T
    T *t = new T();
    SFString name = t->getTypeName();
    SFComponent *TComponent = t->getType()->getComponent();
    //checking if the node belongs to the component.
    if (TComponent != component)
    {
      SFString sg1, sg2;
      if (component->getSceneGraphName() != "")
        sg1 = " of the " + component->getSceneGraphName() + " scene graph";
      if (t->getType()->getSceneGraphName() != "")
        sg1 = " of the " + t->getType()->getSceneGraphName() + " scene graph";  

      cx3d << "warning: defineWalkOnFunction for " << name << " which doesn't belong to the " << component->getName() << " component" << sg1 << "," << std::endl;
      cx3d << "         but to the " << t->getType()->getComponentName() << " component" << sg2 << std::endl;    
    }
    else
      _walkonDict.insert(std::pair<SFString, WalkOnFunction *>(t->getType()->getEncodedName(), WF));
      
    delete t;
    SFType::endFunctionDefinition();
  };
  
  /// Define a new leave function.
  template<class C, class T> 
  void defineLeaveFunction(void (C::*ptrF)(T *) const)
  {
    SFType::beginFunctionDefinition();
    LeaveFunction *LF = new LeaveFunction(reinterpret_cast<ptrToLeaveFunction>(ptrF), this);
    //finding if an occurence of T
    T *t = new T();
    SFString name = t->getTypeName();
    SFComponent *TComponent = t->getType()->getComponent();
    //checking if the node belongs to the component.
    if (TComponent != component)
    {
      SFString sg1, sg2;
      if (component->getSceneGraphName() != "")
        sg1 = " of the " + component->getSceneGraphName() + " scene graph";
      if (t->getType()->getSceneGraphName() != "")
        sg1 = " of the " + t->getType()->getSceneGraphName() + " scene graph";  

      cx3d << "warning: defineLeaveFunction for " << name << " which doesn't belong to the " << component->getName() << " component" << sg1 << "," << std::endl;
      cx3d << "         but to the " << t->getType()->getComponentName() << " component" << sg2 << std::endl;    
    }
    else
      _leaveDict.insert(std::pair<SFString, LeaveFunction *>(t->getType()->getEncodedName(), LF));
      
    delete t;
    SFType::endFunctionDefinition();
  };
  
private:
  EnterDict _enterDict;  
  WalkOnDict _walkonDict;  
  LeaveDict _leaveDict;  
};

}

#endif
